/*
 * rcu.c: Remote Control Unit driver for the Video Disk Recorder
 *
 * Copyright (C) 2000 Klaus Schmidinger
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 * Or, point your browser to http://www.gnu.org/copyleft/gpl.html
 * 
 * The author can be reached at kls@cadsoft.de
 *
 * The project's page is at http://www.cadsoft.de/people/kls/vdr
 *
 * $Id: rcu.c 1.2 2000/02/26 14:16:43 kls Exp $
 */

#case
#zero_ram
#include <16c84.h>
#include <ctype.h>
#fuses   HS,NOPROTECT,NOWDT
#use delay(clock = 10000000)

#define bool short
#define bit  short

// --- Port definitions:

struct { int data  : 5; } port_a;
struct { int addr  : 2;
         int write : 1;
         int txd   : 1;
         int mode  : 1;
         int led   : 1;
         int ir    : 1;
         int rxd   : 1; } port_b;

#define MODE  PIN_B4

#byte port_a = 5
#byte port_b = 6
#byte PORT_A = 5
#byte PORT_B = 6

// --- LED display:

void WriteDigit(char c)
{
  // Writes one digit to the display controller.
  // c    = 0aapdddd
  // aa   = two bit address
  // p    = decimal point
  // dddd = four bit digit code
  port_b.addr = ~(c >> 5);
  port_a.data = c ^ 0x10; // let's have '1' show the decimal point
  port_b.write = 0;
  // implicit 200ns delay
  port_b.write = 1;
}

void Display(long d)
{
  int i;

  for (i = 0; i < 4; i++) {
      WriteDigit((i << 5) | (d & 0x000F));
      d >>= 4;
      }
}

#define MODEb  output_float(MODE)
#define MODEh  output_high(MODE)
#define MODEs  output_low(MODE)

#define LEDon  port_b.led = 0
#define LEDoff port_b.led = 1
#define LEDtog port_b.led ^= 1  // just for debugging

// --- IrDA interface:

char ch;
bool ByteReceived = FALSE;

#int_rb
void GetIrdaByte(void)
{
  int i;

  if (port_b.rxd) {
     ch = 0x00;
     for (i = 8; i--;) {
         // detect one bit:
         delay_us(99); // actually it's 104us, but the rest is loop overhead
         ch >>= 1;
         if (!port_b.rxd)
            ch |= 0x80; // lsb comes in first
         }
     delay_us(100); // wait for stop bit
     if (!port_b.rxd)  // must be '1' (i.e. low signal)
        ByteReceived = TRUE;
     }
}

void PutIrdaByte(char ch)
{
  int i;

  disable_interrupts(RB_CHANGE);
  port_b.txd = 1;
  delay_us(20); // start bit
  port_b.txd = 0;
  for (i = 8; i--;) {
      delay_us(77); // actually it's 84us, but the rest is loop overhead
      port_b.txd = ~ch & 1;
      delay_us(20);
      port_b.txd = 0;
      ch >>= 1;
      }
  enable_interrupts(RB_CHANGE);
  delay_us(124); // stop bit
}

// --- Wait for IR level with timeout

signed long Wait(bit Level, signed long Timeout)
{
  signed long t;
  int i;

  t = Timeout;
  for (; Timeout > 0;) {
      if (port_b.ir == Level) {
         return t - Timeout;
         }
      Timeout -= 10;
      }
  return -1;
}

// --- IR remote control decoder:

enum eIrCode { icNone, // no default - PC must select mode!
               icA,   // 12 bit (SONY)
               icB,   // 16 bit (d-box)
               icC,   // 32 bit (KATHREIN, MASPRO, TechniSat, YAMAHA)
               icD    // 48 bit (Panasonic)
             };

eIrCode IrCode = icNone; // PC must tell us what to do!

#define MAXIRBYTES 6

union {
  char a[MAXIRBYTES];
  struct { long a, b, c; } L; // used for debugging
  } RcCode;

bool b; // can't be inside ReceiveIR() because the it wouldn't use a bit!

bool ReceiveIR(void)
{
  int i;

  memset(&RcCode, 0, sizeof(RcCode));
  b = 0;

#define WAIT(level, max)    (Wait(level, max * 2) >= 0)
#define EXPECT(level, time) (Wait(level, time * 2) > (time / 10 * 9))
#define SYNC(low, high)     (port_b.ir == 1 && WAIT(0, 5000) && EXPECT(1, low) && EXPECT(0, high))

  switch (IrCode) {
    case icA: if (SYNC(2500, 500)) {
                 LEDon;
                 for (i = 12; i--;) {
                     if (!WAIT(0, 500))
                        goto timeout;
                     shift_left(&RcCode, sizeof(RcCode), EXPECT(1, 1250));
                     }
                 }
              else
                 return FALSE;
              break;
    case icB: if (SYNC(600, 2400)) {
                 LEDon;
                 for (i = 16; i--;) {
                     if (!WAIT(!b, 500))
                        goto timeout;
                     shift_left(&RcCode, sizeof(RcCode), b);
                     if (EXPECT(b, 900))
                        b = !b;
                     }
                 }
              else
                 return FALSE;
              break;
    case icC: if (SYNC(9000, 4500)) {
                 LEDon;
                 for (i = 32; i--;) {
                     if (!WAIT(1, 600))
                        goto timeout;
                     shift_left(&RcCode, sizeof(RcCode), EXPECT(0, 1600));
                     }
                 }
              else
                 return FALSE;
              break;
    case icD: if (SYNC(3600, 1700)) {
                 LEDon;
                 for (i = 48; i--;) {
                     if (!WAIT(1, 500))
                        goto timeout;
                     shift_left(&RcCode, sizeof(RcCode), EXPECT(0, 1200));
                     }
                 }
              else
                 return FALSE;
              break;
    default:  return FALSE; // no code selected
    }
  LEDoff;
  return !ByteReceived; // any serial interrupts might have messed up the data
timeout:
  LEDoff;
  return FALSE;
}

void SendIR(void)
{
  int i;

  PutIrdaByte('X');
  for (i = MAXIRBYTES; i--;)
      PutIrdaByte(RcCode.a[i]);
}

void main(void)
{
  // Initialize ports:
  set_tris_a(0b11100000);
  set_tris_b(0b11000000);

  LEDoff;
  MODEb;
  Display(0xCBD0); // a friendly "HELO"

  enable_interrupts(RB_CHANGE);
  enable_interrupts(GLOBAL);
  for (;;) {
      if (ByteReceived) {
         if ((ch & 0x80) == 0) // msb == 1 indicates command
            WriteDigit(ch);
         else {
            switch (ch & 0x7F) {
              case '0': LEDoff; break;
              case '1': LEDon;  break;
              case 'A':
              case 'B':
              case 'C':
              case 'D': IrCode = (ch & 0x7F) - 'A' + 1; break;
              case 'b': MODEb;  break;
              case 'h': MODEh;  break;
              case 's': MODEs;  break;
              }
            }
         ByteReceived = FALSE;
         PutIrdaByte(ch); // send it back as handshake
         }
      if (ReceiveIR()) {
//         Display(RcCode.L.a);  // display RC code for debugging
         SendIR();
         }
      }
}
